import logging
import re
import warnings
from contextlib import contextmanager
from enum import Enum

import crayons
from dateutil import parser
from packaging import version
from pbr.version import VersionInfo
from pytz import UTC

VERSION = VersionInfo('flask-restless-client').release_string()
RECOMMENDED_SERVER_VERSION = ['0.2.1', '0.2.2', '0.2.3']

LIKELY_PARSABLE_DATETIME = r"^(\d{4}-\d{2}-\d{2}[ T]\d{2}:\d{2}:\d{2})|(\d{8}T\d{6}Z?)"
logger = logging.getLogger('restless-client')

LOCAL_ID_COUNT = 0
DEPTH = 0


class State(Enum):
    VOID = 1
    LOADING = 2
    LOADABLE = 3


def generate_id():
    global LOCAL_ID_COUNT
    LOCAL_ID_COUNT += 1
    return 'C{}'.format(LOCAL_ID_COUNT)


def urljoin(*args):
    args = [a.strip('/') for a in args]
    return "/".join(args)


def datetime_from_value(value, as_timezone):
    if isinstance(value, str) and re.search(LIKELY_PARSABLE_DATETIME, value):
        return parser.parse(value).astimezone(as_timezone)
    elif isinstance(value, list):
        for idx, item in enumerate(value):
            if isinstance(item, dict):
                value[idx] = parse_custom_types(item)
            else:
                value[idx] = datetime_from_value(item, as_timezone)

    return value


def parse_custom_types(dct, as_timezone=UTC, **kwargs):
    for k, v in dct.items():
        try:
            if isinstance(v, dict):
                dct[k] = parse_custom_types(v, as_timezone)
            else:
                dct[k] = datetime_from_value(v, as_timezone)
        except Exception:
            pass

    return dct


class UserException(Exception):
    pass


class LoadingManager:
    def __init__(self, client):
        self.client = client
        self.active_contexts = 0

    def __enter__(self):
        self.active_contexts += 1
        self.client.state = State.LOADING

    def __exit__(self, exc_type, exc_value, traceback):
        self.active_contexts -= 1
        if self.active_contexts == 0:
            self.client.state = State.LOADABLE


class RelationHelper:
    def __init__(self, client, opts, relations):
        self.client = client
        self.opts = opts
        self.relations = relations

    def type(self, name):  # noqa A003
        return self.relations[name]['relation_type']

    def model_name(self, name):
        return self.relations[name]['foreign_model']

    def model(self, name):
        return self.client._classes[self.model_name(name)]

    def column_name(self, name):
        if name not in self.relations:
            return name
        return self.relations[name].get('local_column', name)

    def backref(self, name):
        return self.relations[name].get('backref')

    def is_scalar(self, name):
        return self.type(name) in ('MANYTOONE', 'ONETOONE')

    def is_valid_instance(self, name, instance):
        if name in self.relations.keys() and self.is_scalar(name):
            allowed = (self.model(name), self.opts.LoadableProperty,
                       type(None))
            if not isinstance(instance, allowed):
                msg = '{} must be an instance of {}, not {}'
                raise Exception(
                    msg.format(name, self.model_name(name),
                               instance.__class__.__name__))


@contextmanager
def pretty_logger(depth=2):
    global DEPTH
    DEPTH += depth
    yield
    DEPTH -= depth


def get_depth():
    return DEPTH


def check_server_compatibility(server_version):
    message = ("\n\nYour {0} version is more recent than the {1} version, "
               "incompatibilities may arise.\nConsider downgrading your {0} "
               "version or upgrading your {1} version.\n\nServer version: {2}"
               "\nRecommended server version: {3}\n\n")

    params = ('client', 'server', 'N/A', RECOMMENDED_SERVER_VERSION)
    if server_version:
        server_version = version.parse(server_version)
        recommended = [version.parse(v) for v in RECOMMENDED_SERVER_VERSION]

        if server_version in recommended:
            return

        if server_version < min(recommended):
            params = ('client', 'server', server_version, min(recommended))
        if server_version > max(recommended):
            params = ('server', 'client', server_version, max(recommended))

    warnings.warn(str(crayons.yellow(message.format(*params))))



import logging

from restless_client.filter import ComparisonResult, FilterMixIn
from restless_client.utils import State

logger = logging.getLogger('restless-client')
NOT_UPDATING_MSG = "Not updating {}.{} (dirty and in load state)"


class LoadableProperty:
    def __init__(self, attribute):
        self.attribute = attribute

    def __set__(self, obj, value):
        dirty = self.attribute in obj._rlc.dirty
        if dirty and obj._rlc.client.is_loading:
            logger.debug(NOT_UPDATING_MSG.format(obj, self.attribute))
            return
        obj._rlc.values[self.attribute] = value
        if not obj._rlc.client.is_loading:
            obj._rlc.dirty.add(self.attribute)

    def __get__(self, obj, objtype=None):
        if objtype and obj is None:
            # if obj is None, the __get__ is invoked on class level
            # returning FilterNode will allow the user to build filters
            return FilterNode(objtype, self.attribute)
        if self.getval(obj) is State.VOID and not obj._rlc.is_new:
            args = (obj.__class__, obj._rlc.pk_val)
            logger.debug('Loading {} with id {} remotely'.format(*args))
            obj._rlc.connection.load(*args)
        if self.getval(obj) is State.VOID:
            return None
        return self.getval(obj)

    def getval(self, obj):
        return obj._rlc.values.get(self.attribute, State.VOID)


class FilterNode(FilterMixIn):
    def __init__(self, parent, parent_attribute, parent_node=None):
        self.parent_klass = parent
        self.parent_attribute = parent_attribute
        self.parent_node = parent_node
        self.klass = None
        self.rel_type = None
        self.is_leaf = True
        if parent_attribute in parent._rlc._relations:
            self.rel_type = parent._rlc._relations[parent_attribute][
                'relation_type']
            class_name = parent._rlc._relations[parent_attribute][
                'foreign_model']
            self.klass = parent._rlc.client._classes[class_name]

    def __getattr__(self, attr):
        meta = self.klass._rlc
        if attr not in meta.attributes() and attr not in meta.relations():
            msg = '{} has no attribute named {}'
            raise AttributeError(msg.format(meta.class_name, attr))
        return FilterNode(self.klass, attr, parent_node=self)

    def _assemble_filter(self, op, val):
        attribute, val = self._transform_relation(self.parent_attribute, val)

        rfilter = ComparisonResult(attribute, op, self._clean(val))
        if not self.is_leaf:
            return rfilter

        rp = self
        while rp.parent_node:
            rp.parent_node.is_leaf = False
            if rp.parent_node.rel_type == 'ONETOMANY':
                rfilter = rp.parent_node.any_(rfilter)
            else:
                rfilter = rp.parent_node.has_(rfilter)
            rp = rp.parent_node
        return rfilter

    def _transform_relation(self, attribute, val):
        if hasattr(val, '__class__') and hasattr(val.__class__, '__bases__'):
            for klass in val.__class__.__bases__:
                if klass.__name__ == 'BaseObject':
                    val = val._rlc.pk_val
                    rel_def = self.parent_klass._rlc._relations[
                        self.parent_attribute]
                    attribute = rel_def['local_column']
        return attribute, val



from unittest import mock

import pytest

from restless_client.filter import ComparisonResult
from restless_client.inspect import InstanceState, ModelMeta, inspect
from restless_client.property import FilterNode, LoadableProperty


@pytest.fixture
def Stub():
    client = mock.Mock(is_loading=False, connection=mock.Mock())

    class StubClass:
        _relations = {}
        oid = 1
        attribute1 = LoadableProperty('attribute1')
        _rlc = ModelMeta(client=client,
                         relations=[],
                         pk_name='oid',
                         class_name=None,
                         attributes=None,
                         properties=None,
                         methods=None,
                         base_url=None,
                         method_url=None,
                         property_url=None,
                         polymorphic=None,
                         relhelper=None)

        def __init__(self, **kwargs):
            self._rlc.client = kwargs.get('client', client)
            self._rlc = InstanceState(self)
            self._rlc.dirty = kwargs.get('dirty', set())
            self._rlc.values = kwargs.get('values', {})

    return StubClass


def test_it_can_set_an_attribute(Stub):
    stub = Stub()
    stub.attribute1 = 'test'
    assert stub.attribute1 == 'test'
    assert Stub().attribute1 != 'test'


def test_it_can_load_an_attribute(Stub):
    stub = Stub()
    assert stub.attribute1 == None
    inspect(stub).connection.load.assert_called_once()


def test_it_can_return_loaded_values(Stub):
    stub = Stub()
    inspect(stub).values['attribute1'] = 'test'
    assert stub.attribute1 == 'test'


def test_it_flags_attribute_as_dirty_when_set(Stub):
    stub = Stub()
    stub.attribute1 = 'test'
    assert 'attribute1' in inspect(stub).dirty


def test_it_does_not_update_when_dirty_and_loading(Stub):
    client = mock.Mock(is_loading=True)
    stub = Stub(client=client)
    inspect(stub).dirty.add('attribute1')
    inspect(stub).values['attribute1'] = 'test'
    stub.attribute1 = 'not test'
    assert stub.attribute1 == 'test'


def test_it_returns_a_filter_as_class_attribute(Stub):
    assert type(Stub.attribute1 == 1) == ComparisonResult


def test_is_returns_relation_property(Stub):
    client = mock.Mock(_classes={'SOMEMODEL': 'SOMEMODEL'})
    Stub._client = client
    inspect(Stub).relations = {
        'attribute1': {
            'relation_type': 'WHOTOCARES',
            'foreign_model': 'SOMEMODEL'
        }
    }
    assert type(Stub.attribute1) == FilterNode



import os
import time
from contextlib import contextmanager
from datetime import date, datetime, timedelta
from multiprocessing import Process

import cereal_lazer
import flask_restless
import pytest
from fast_alchemy import FlaskFastAlchemy
from flask import Flask
from flask_restless_datamodel import DataModel
from flask_sqlalchemy import SQLAlchemy
from requests_flask_adapter import Session
from sqlalchemy.ext.hybrid import hybrid_property

from restless_client import Client
from restless_client.ext.auth import BaseSession

ROOT_DIR = os.path.dirname(__file__)
DATA_DIR = os.path.join(ROOT_DIR, 'data')
API_METHODS = ['GET', 'PUT', 'POST', 'DELETE']


class RaiseSession(BaseSession, Session):
    def request(self, *args, **kwargs):
        return super().request(*args, **kwargs)


def build_endpoints(app, fa):
    manager = flask_restless.APIManager(app, flask_sqlalchemy_db=app.db)
    for class_name, class_ in fa.class_registry.items():
        manager.create_api(class_, methods=API_METHODS)
        setattr(app, class_name, class_)
    data_model = DataModel(manager)
    manager.create_api(data_model, methods=['GET'])
    app.manager = manager


@pytest.fixture(scope='function')
def app():
    app = Flask(__name__)
    app.config['SQLALCHEMY_DATABASE_URI'] = "sqlite:///:memory:"
    app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False
    app.config['PROPAGATE_EXCEPTIONS'] = True
    db = SQLAlchemy(app)
    app.db = db
    return app


@pytest.fixture(scope='function')
def session(app):
    return app.db.session


@pytest.fixture(scope='function')
def instances(app):
    fa = FlaskFastAlchemy(app.db)
    fa.load(os.path.join(DATA_DIR, 'instances.yaml'))
    build_endpoints(app, fa)
    return fa


@pytest.fixture(scope='function')
def filters(app):
    fa = FlaskFastAlchemy(app.db)
    fa.load(os.path.join(DATA_DIR, 'filter.yaml'))
    build_endpoints(app, fa)
    return app


@pytest.fixture
def cl(app, instances):
    RaiseSession.register('http://app', app)
    return Client(url='http://app', session=RaiseSession(), debug=True)


@pytest.fixture
def fcl(filters):
    RaiseSession.register('http://app', filters)
    return Client(url='http://app', session=RaiseSession(), debug=True)


@pytest.fixture
def mcl(app, session, instances):
    db = app.db

    class MisterTyper(db.Model):
        __tablename__ = "mister_typer"
        id = db.Column(db.Integer, primary_key=True)
        date = db.Column(db.Date())
        dt = db.Column(db.DateTime())
        json = db.Column(db.JSON())
        interval = db.Column(db.Interval())
        binary = db.Column(db.Binary())
        num = db.Column(db.Numeric())

    db.create_all()

    class Apartment(instances.Formicarium):
        __mapper_args__ = {'polymorphic_identity': 'apartment'}

        @property
        def some_property(self):
            return 'a_property_value'

        @property
        def settable_property(self):
            return self.name

        @settable_property.setter
        def settable_property(self, value):
            self.name == value

        @hybrid_property
        def some_hybrid(self):
            return self.collection

        def function_without_params(self):
            return 5

        def function_with_params(self, arg1, arg2):
            return '{}: {}'.format(arg1, arg2)

        def function_with_kwargs(self, kwarg1, kwarg2):
            return self.function_with_params(kwarg1, kwarg2)

        def funtion_with_args_kwargs(self, arg1, kwarg1):
            return self.function_with_params(arg1, kwarg1)

        def function_with_default_params(self, param1=5, param2=6):
            return self.function_with_params(param1, param2)

        def function_with_an_object(self, obj):
            assert isinstance(obj, app.AntColony)
            return obj

        def function_with_uncommitted_object(self):
            return Apartment(name='Oof-Owie')

        def function_with_new_obj(self):
            mt = MisterTyper(
                date=date(2018, 1, 1),
                dt=datetime(2018, 1, 1),
                json={"awe": "some"},
            )
            session.add(mt)
            session.commit()
            return mt

    Apartment.__tablename__ = 'apartment'

    apt = Apartment(name='ApAntMent')
    apt.collection = instances.AntCollection.query.first()
    session.add(apt)
    session.commit()

    app.manager.create_api(Apartment, methods=API_METHODS)
    app.manager.create_api(MisterTyper, methods=API_METHODS)

    RaiseSession.register('http://app', app)

    yield Client(url='http://app', session=RaiseSession(), debug=True)



import logging

import crayons

from .inspect import inspect
from .utils import generate_id

logger = logging.getLogger('restless-client')


def get_class(cls, kwargs, meta):
    if not meta.polymorphic.get('identities'):
        return cls
    discriminator_name = meta.polymorphic['on']
    for discriminator, kls in meta.polymorphic['identities'].items():
        if kwargs.get(discriminator_name) == discriminator:
            return meta.client._classes[kls]
    return cls


class BaseObject:
    def __init__(self, **kwargs):
        oid = self._rlc.pk_name
        super().__setattr__(oid,
                            kwargs[oid] if oid in kwargs else generate_id())
        self._rlc.deserializer.load(self, kwargs)
        self._rlc.client._register(self)

    def __new__(cls, **kwargs):
        key = None
        meta = cls._rlc
        if kwargs.get(meta.pk_name):
            key = '%s%s' % (cls.__name__, kwargs[meta.pk_name])
        if key in meta.client.registry:
            obj = meta.client.registry[key]
            logger.debug(crayons.yellow('Using existing {}'.format(key)))
        else:
            cls = get_class(cls, kwargs, meta)
            obj = object.__new__(cls)
            obj._rlc = inspect(obj)
            logger.debug(crayons.yellow('initialising {}'.format(key)))
        return obj

    def __setattr__(self, name, value):
        if not name.startswith('_'):
            if name not in self._rlc.settable_attributes:
                raise AttributeError('{} has no attribute named {}'.format(
                    self._rlc.class_name, name))
        self._rlc.relhelper.is_valid_instance(name, value)
        object.__setattr__(self, name, value)

    def __repr__(self):
        return str(self)

    def __str__(self):
        attributes = ["{}: {}".format(self._rlc.pk_name, self._rlc.pk_val)]
        if hasattr(self, 'name'):
            attributes.append("name: {}".format(self.name))
        return "<{} [{}]>".format(self.__class__.__name__,
                                  " | ".join(attributes))



import logging
from datetime import date, datetime
from itertools import chain

import crayons

from .types import cast_type, object_hook_emit
from .utils import State, pretty_logger

logger = logging.getLogger('restless-client')
LOAD_MSG = 'loading {}.{} with value {}'


def log(o, a, v, attr_color='red'):
    logger.info(
        LOAD_MSG.format(
            crayons.yellow(o.__class__.__name__, always=True, bold=True),
            getattr(crayons, attr_color)(a, always=True, bold=True),
            crayons.green(str(v)[:150], always=True, bold=True)))


def log_loading(color):
    def decorator(fn):
        def execute(self, obj, field, val, *args, **kwargs):
            log(obj, field, val, color)
            return fn(self, obj, field, val, *args, **kwargs)

        return execute

    return decorator


def set_attr(obj, attr, value):
    if value and value is not State.VOID:
        value = object_hook_emit(obj.__class__, attr, value)
    setattr(obj, attr, value)


class ObjectSerializer:
    def __init__(self, client, opts):
        self.client = client
        self.opts = opts

    def serialize(self, obj):
        to_serialize = list(chain(obj._rlc.attributes(), obj._rlc.relations()))
        return self._serialize(obj, to_serialize)

    def serialize_dirty(self, obj):
        return self._serialize(obj, obj._rlc.dirty, autosave=True)

    def _serialize(self, obj, to_serialize, autosave=False):
        # create attribute dict
        if obj._rlc.pk_name in to_serialize:
            to_serialize.remove(obj._rlc.pk_name)
        return self._raw_serialize(obj, to_serialize, autosave)

    def _raw_serialize(self, obj, to_serialize, autosave=False):
        object_dict = {}
        for attr in to_serialize:
            if attr not in obj._rlc.values:
                continue
            value = obj._rlc.values[attr]
            object_dict[attr] = self.clean(attr, value, autosave)
        return object_dict

    def clean(self, attr, value, autosave):
        if isinstance(value, self.opts.BaseObject):
            if value._rlc.is_new and autosave:
                value._rlc.save()
            value = {value._rlc.pk_name: value._rlc.pk_val}
        if isinstance(value, (date, datetime)):
            value = value.isoformat()
        if isinstance(value, (list, set, tuple)):
            value = [self.clean(attr, v, autosave) for v in value]
        return value


class ObjectDeserializer:
    def __init__(self, client, opts):
        self.client = client
        self.opts = opts

    def load(self, obj, attributes_dict):
        with pretty_logger():
            self.handle_attributes(obj, attributes_dict)
            self.handle_relations(obj, attributes_dict)

    def handle_attributes(self, obj, raw):
        attrs = obj._rlc._attributes
        for field, field_type in attrs.items():
            if field in raw:
                log(obj, field, raw[field])
                value_type = attrs.get(field)
                value = cast_type(raw[field], value_type)
                set_attr(obj, field, value)

    def handle_relations(self, obj, raw):
        relation_type_handlers = {
            'ONETOMANY': self.handle_o2m,
            'MANYTOONE': self.handle_m2o,
            'ONETOONE': self.handle_o2o,
            'MANYTOMANY': self.handle_o2m,
        }
        for field in obj._rlc._relations.keys():
            val = raw.get(field, State.VOID)
            handler = relation_type_handlers[obj._rlc.relhelper.type(field)]
            handler(obj, field, val, obj._rlc.relhelper.model(field))

    @log_loading('blue')
    def handle_o2m(self, obj, field, val, rel_model):
        typed_list = self.opts.TypedListClass(rel_model, obj, field)
        if val is not State.VOID:
            for rel_obj in val:
                if isinstance(rel_obj, dict):
                    rel_obj = rel_model(**rel_obj)
                with pretty_logger():
                    typed_list.append(rel_obj)
            set_attr(obj, field, typed_list)
        elif obj._rlc.is_new:
            set_attr(obj, field, typed_list)

    @log_loading('cyan')
    def handle_m2o(self, obj, field, val, rel_model):
        if isinstance(val, dict):
            val = rel_model(**val)
        if hasattr(val.__class__, '__bases__'):
            if self.opts.BaseObject in val.__class__.__bases__:
                set_attr(obj, field, val)

    @log_loading('magenta')
    def handle_o2o(self, obj, field, val, rel_model):
        self.handle_m2o(obj, field, val, rel_model)



import pytest
from requests.exceptions import HTTPError

from restless_client.filter import FilterMixIn


def test_eq(fcl):
    o = fcl.Object1
    res = o.query.filter(o.attribute2 == 4).all()
    assert len(res) == 1
    assert res[0].id == 4 and res[0].attribute2 == 4


def test_ne(fcl):
    o = fcl.Object1
    res = o.query.filter(o.attribute2 != 4).all()
    assert len(res) == 4
    assert sorted([ob.id for ob in res]) == [1, 2, 3, 5]
    assert sorted([ob.attribute2 for ob in res]) == [1, 2, 3, 5]


def test_invert(fcl):
    o = fcl.Object1
    res = o.query.filter(~(o.attribute2 == 4)).all()
    assert len(res) == 4
    assert sorted([ob.id for ob in res]) == [1, 2, 3, 5]
    assert sorted([ob.attribute2 for ob in res]) == [1, 2, 3, 5]


def test_lt(fcl):
    o = fcl.Object1
    res = o.query.filter(o.attribute2 < 4).all()
    assert len(res) == 3
    assert sorted([ob.id for ob in res]) == [1, 2, 3]
    assert sorted([ob.attribute2 for ob in res]) == [1, 2, 3]


def test_lte(fcl):
    o = fcl.Object1
    res = o.query.filter(o.attribute2 <= 4).all()
    assert len(res) == 4
    assert sorted([ob.id for ob in res]) == [1, 2, 3, 4]
    assert sorted([ob.attribute2 for ob in res]) == [1, 2, 3, 4]


def test_gt(fcl):
    o = fcl.Object1
    res = o.query.filter(o.attribute2 > 4).all()
    assert len(res) == 1
    assert sorted([ob.id for ob in res]) == [5]
    assert sorted([ob.attribute2 for ob in res]) == [5]


def test_gte(fcl):
    o = fcl.Object1
    res = o.query.filter(o.attribute2 >= 4).all()
    assert len(res) == 2
    assert sorted([ob.id for ob in res]) == [4, 5]
    assert sorted([ob.attribute2 for ob in res]) == [4, 5]


def test_in(fcl):
    o = fcl.Object1
    res = o.query.filter(o.attribute2.in_([1, 2, 3])).all()
    assert len(res) == 3
    assert sorted([ob.id for ob in res]) == [1, 2, 3]
    assert sorted([ob.attribute2 for ob in res]) == [1, 2, 3]


def test_mix_of_things(fcl):
    o = fcl.Object1
    res = o.query.filter(o.attribute2.in_([2, 3, 4, 5]), o.id > 2,
                         o.id < 5).all()
    assert len(res) == 2
    assert sorted([ob.id for ob in res]) == [3, 4]
    assert sorted([ob.attribute2 for ob in res]) == [3, 4]


def test_single_fails_on_multi_result(fcl):
    o = fcl.Object1
    with pytest.raises(Exception):
        o.query.filter(o.attribute2.in_([1, 2, 3])).one()


def test_single(fcl):
    o = fcl.Object1
    res = o.query.filter(o.attribute2 == 4).one()
    assert res.id == 4 and res.attribute2 == 4


def test_one_or_none(fcl):
    o = fcl.Object1
    res = o.query.filter(o.attribute2 == 99).one_or_none()
    assert res == None


def test_one_or_none_fails_on_multiple(fcl):
    o = fcl.Object1
    with pytest.raises(Exception):
        o.query.filter(o.attribute2.in_([1, 2, 3])).one_or_none()


def test_filter_by(fcl):
    o = fcl.Object1
    res = o.query.filter_by(attribute2=4).all()
    assert len(res) == 1
    assert res[0].id == 4 and res[0].attribute2 == 4


def test_raw_filter():
    f = FilterMixIn()
    f.attribute = 'test'
    expected = {'name': 'test', 'op': '!=', 'val': 2}
    assert (~(f == 2)).to_raw_filter() == expected


def test_complex_filter():
    f = FilterMixIn()
    f.attribute = 'test'
    expected = {
        'or': [{
            'and': [{
                'name': 'test',
                'op': 'has',
                'val': {
                    'name': 'test',
                    'op': '==',
                    'val': 2
                }
            }, {
                'name': 'test',
                'op': 'not_in',
                'val': [1, 2, 3]
            }]
        }, {
            'name': 'test',
            'op': '>=',
            'val': 5
        }]
    }
    res = (f.has_((f == 2)) & (~f.in_([1, 2, 3])) | (f >= 5)).to_raw_filter()
    assert res == expected


def test_o2m_relation_filter(fcl):
    o = fcl.Object2
    f = o.relation1.relation2.attribute1 == "o3a11"
    expected = {
        'name': 'relation1',
        'op': 'has',
        'val': {
            'name': 'relation2',
            'op': 'has',
            'val': {
                'name': 'attribute1',
                'op': '==',
                'val': 'o3a11'
            }
        }
    }
    assert f.to_raw_filter() == expected


def test_m2o_relation_filter(fcl):
    o = fcl.Object3
    f = o.relation2.relation1.attribute1 == "o2a13"
    expected = {
        'name': 'relation2',
        'op': 'any',
        'val': {
            'name': 'relation1',
            'op': 'any',
            'val': {
                'name': 'attribute1',
                'op': '==',
                'val': 'o2a13'
            }
        }
    }
    assert f.to_raw_filter() == expected


def test_mix_relation_filter(fcl):
    o = fcl.Object1
    f = o.relation1.relation1.attribute1 == "o1a11"
    expected = {
        'name': 'relation1',
        'op': 'any',
        'val': {
            'name': 'relation1',
            'op': 'has',
            'val': {
                'name': 'attribute1',
                'op': '==',
                'val': 'o1a11'
            }
        }
    }
    assert f.to_raw_filter() == expected


def test_it_raises_an_error_on_unknown_attribute(fcl):
    with pytest.raises(AttributeError) as exc:
        assert fcl.Object1.relation1.Unknown == None


def test_filter_does_not_prevent_chaining_with_query(fcl):
    o = fcl.Object3
    oneobj = o.query.filter(o.attribute1 == "o3a11").one()
    assert isinstance(oneobj, o)

    allobjs = o.query.all()
    assert all([isinstance(obj, o) for obj in allobjs])



import logging
import sys
from collections import defaultdict
from itertools import chain

import crayons
from cereal_lazer import Cereal

from .collections import ObjectCollection, TypedList
from .connection import Connection
from .ext.auth import Session
from .filter import QueryFactory
from .inspect import ModelMeta
from .marshal import ObjectDeserializer, ObjectSerializer
from .method import Method, construct_method
from .models import BaseObject
from .property import LoadableProperty
from .utils import check_server_compatibility  # noqa
from .utils import LoadingManager, RelationHelper, State, get_depth, urljoin


class UncallableProperty(Exception):
    pass


def register_serializer(model):
    rlc = model._rlc
    to_serialize = list(chain(rlc.attributes(), rlc.relations()))

    def load_model(value):
        return model(**value)

    def serialize_model(value):
        return rlc.serializer._raw_serialize(value, to_serialize)

    rlc.client.cereal.register_class(rlc.class_name, model, serialize_model,
                                     load_model)


class DepthFilter(logging.Filter):
    def filter(self, record):  # noqa A003
        record.depth = crayons.yellow('{}>'.format('-' * get_depth()), True,
                                      True)
        return True


logger = logging.getLogger('restless-client')
logger.addFilter(DepthFilter())

steam_handler = logging.StreamHandler(sys.stdout)
steam_handler.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(depth)s %(message)s')
steam_handler.setFormatter(formatter)
logger.addHandler(steam_handler)
logger.propagate = False


class Options:
    def __init__(self, opts):
        # the only entrypoint to load/save/update objects
        self.ConnectionClass = opts.pop('connection', Connection)
        # takes care of populating instances based on the given kwargs
        self.DeserializeClass = opts.pop('deserializer', ObjectDeserializer)
        self.SerializeClass = opts.pop('serializer', ObjectSerializer)
        self.RelationHelper = opts.pop('relhelper', RelationHelper)
        # takes care of building the classes
        self.ConstructorClass = opts.pop('constructor', ClassConstructor)
        # base object every constructed class will inherit from
        self.BaseObject = opts.pop('base_object', BaseObject)
        # type of list used to group multiple objects when executing a query
        self.CollectionClass = opts.pop('collection_class', ObjectCollection)
        # type of list used to keep track of instance relations
        self.TypedListClass = opts.pop('typed_list', TypedList)
        # the property used by constructed classes to handle model attributes
        self.LoadableProperty = opts.pop('loadable_property', LoadableProperty)
        # how to reach the server when calling an object function
        self.Method = opts.pop('method_class', Method)
        self.ServerProperty = opts.pop('server_property_class', ServerProperty)

        self.debug = opts.pop('debug', True)
        self.data_model_endpoint = opts.pop('data_model_endpoint',
                                            'api/flask-restless-datamodel')

        # cereal lazer options
        # will return the raw data instead of raising an error when loading
        self.raise_load_errors = opts.pop('raise_load_errors', True)
        # will try to coerce non registered classes into an emulated object
        self._serialize_naively_set_by_user = 'serialize_naively' in opts
        self.serialize_naively = opts.pop('serialize_naively', False)

        if 'session' in opts:
            self.session = opts.pop('session')
        else:
            auth_url = opts.pop('auth_url',
                                urljoin(opts['base_url'], 'api/auth'))
            self.session = Session(auth_url, **opts)


class ServerProperty:
    def __init__(self, attribute, connection):
        self.attribute = attribute
        self.connection = connection

    def __get__(self, obj, objtype=None):
        if objtype and obj is None:
            return self

        if obj._rlc.is_new:
            raise UncallableProperty("Cannot call properties on new objects.")

        url = self.get_url(obj)
        result = self.connection.request(url, http_method='get')
        return self.cereal.loads(result['payload'])

    def get_url(self, obj):
        rlc = obj._rlc
        return '{}/{}/{}'.format(rlc.property_url, rlc.pk_val, self.attribute)

    @property
    def cereal(self):
        return self.connection.client.cereal


class SettableServerProperty(ServerProperty):
    def __init__(self, attribute, connection):
        super().__init__(attribute, connection)
        self.value = defaultdict(lambda: State.VOID)

    def __get__(self, obj, objtype=None):
        if self.value[obj] == State.VOID:
            return super().__get__(obj, objtype)
        return self.value[obj]

    def __set__(self, obj, value):
        self.value[obj] = value
        obj._rlc.dirty.add(self.attribute)

    def _commit(self, obj):
        if self.value[obj] == State.VOID:
            return
        url = self.get_url(obj)
        payload = self.cereal.dumps(self.value[obj])
        self.connection.request(url, http_method='post', json=payload)


class ClassConstructor:
    def __init__(self, client, opts):
        self.client = client
        self.opts = opts

    def construct_class(self, name, details):
        base_url = urljoin(
            self.client.base_url,
            details['url_prefix'],
        )

        meta = ModelMeta(
            client=self.client,
            class_name=name,
            pk_name=details['pk_name'],
            attributes=details['attributes'],
            properties=details['properties'],
            relations=details['relations'],
            methods=details['methods'].keys(),
            base_url=urljoin(base_url, details['collection_name']),
            method_url=urljoin(base_url, 'method', name.lower()),
            property_url=urljoin(base_url, 'property', name.lower()),
            polymorphic=details.get('polymorphic', {}),
            relhelper=self.opts.RelationHelper(self.client, self.opts,
                                               details['relations']))
        attributes = {'_rlc': meta}
        for field in chain(details['attributes'], details['relations']):
            attributes[field] = self.opts.LoadableProperty(field)

        if self.opts.ServerProperty:
            for field, url in details['properties'].items():
                attributes[field] = self.opts.ServerProperty(
                    field, self.client.connection)

        if self.opts.Method:
            for method, method_details in details['methods'].items():
                attributes[method] = construct_method(self.opts, self.client,
                                                      method, method_details)

        inherits = [self.opts.BaseObject]
        if details.get('polymorphic', {}).get('parent'):
            parent = self.client._classes[details['polymorphic']['parent']]
            inherits.insert(0, parent)
        klass = type(str(name), tuple(inherits), attributes)
        klass.query = QueryFactory(self.client.connection, klass)
        self.client._classes[name] = klass
        setattr(self.client, name, klass)
        register_serializer(klass)


class Client:
    def __init__(self, url, **kwargs):
        self.base_url = url
        self.state = State.LOADABLE

        self.registry = {}
        self._classes = {}

        kwargs['base_url'] = url
        self.opts = Options(kwargs)

        self.connection = self.opts.ConnectionClass(self, self.opts)
        self.serializer = self.opts.SerializeClass(self, self.opts)
        self.deserializer = self.opts.DeserializeClass(self, self.opts)
        self.constructor = self.opts.ConstructorClass(self, self.opts)
        self.cereal = Cereal(
            serialize_naively=self.opts.serialize_naively,
            raise_load_errors=self.opts.raise_load_errors,
        )
        self.initialize()
        self.__loading_manager = LoadingManager(self)

    def initialize(self):
        url = urljoin(self.base_url, self.opts.data_model_endpoint)
        res = self.connection.request(url)
        meta = res.pop('FlaskRestlessDatamodel', {})
        check_server_compatibility(meta.get('server_version'))
        if not self.opts._serialize_naively_set_by_user:
            self.cereal.serialize_naively = meta['serialize_naively']
        delayed = {}
        for name, details in res.items():
            if details.get('polymorphic', {}).get('parent'):
                delayed[name] = details
                continue
            self.constructor.construct_class(name, details)
        for name, details in delayed.items():
            self.constructor.construct_class(name, details)

    @property
    def loading(self):
        return self.__loading_manager

    @property
    def is_loading(self):
        return self.state == State.LOADING

    def _register(self, obj):
        self.registry['%s%s' % (obj.__class__.__name__, obj._rlc.pk_val)] = obj

    def delete(self, instance):
        instance._rlc.delete()

    def save(self, instance=None):
        if instance is None:
            for obj in self.registry.values():
                if obj._rlc.dirty:
                    obj._rlc.save()
        else:
            instance._rlc.save()



from datetime import datetime
from unittest.mock import Mock

import pytest

from restless_client.inspect import inspect
from restless_client.marshal import ObjectDeserializer, ObjectSerializer
from restless_client.models import BaseObject


class TypedList(list):
    def __init__(self, otype, parent, for_attr=None, *args, **kwargs):
        self.type = otype
        self.parent = parent
        self.for_attr = for_attr
        super().__init__(*args, **kwargs)


@pytest.fixture
def ds():
    return ObjectDeserializer(
        Mock(), Mock(
            TypedListClass=TypedList,
            BaseObject=Mock,
        ))


@pytest.fixture
def s():
    return ObjectSerializer(Mock(), Mock(BaseObject=Mock))


def test_it_can_deserialize_an_attribute(ds):
    expected = {
        'attr1': 'val1',
        'attr2': 'val2',
    }
    obj = Mock(_rlc=Mock(_attributes=expected))
    ds.handle_attributes(obj, expected)
    assert obj.attr1 == 'val1'
    assert obj.attr2 == 'val2'


def test_it_can_deserialize_a_o2m_relation(ds):
    obj = Mock()
    reldef = [{'id': 1, 'attr1': 'someattr'}]
    ds.handle_o2m(obj, 'rel1', reldef, Mock)
    assert obj.rel1[0].id == 1
    assert obj.rel1[0].attr1 == 'someattr'


def test_it_can_deserialize_a_m2o_relation(ds):
    obj = Mock()
    reldef = {'id': 1, 'attr1': 'someattr'}
    ds.handle_m2o(obj, 'rel1', reldef, Mock)
    assert obj.rel1.id == 1
    assert obj.rel1.attr1 == 'someattr'


def test_it_can_deserialize_a_o2o_relation(ds):
    obj = Mock()
    reldef = {'id': 1, 'attr1': 'someattr'}
    ds.handle_o2o(obj, 'rel1', reldef, Mock)
    assert obj.rel1.id == 1
    assert obj.rel1.attr1 == 'someattr'


def setup_obj():
    relhelper = Mock()
    relhelper.column_name.side_effect = lambda x: x
    rlc = Mock(pk_name='id',
               relhelper=relhelper,
               values={
                   'id': 1,
                   'attr1': 'someattr',
                   'somedate': datetime(2018, 1, 1),
                   'somelist': [1, 2, 3],
                   'rel1': Mock(_rlc=Mock(pk_val=2, pk_name='id')),
                   'rel2': [Mock(_rlc=Mock(pk_val=3, pk_name='id'))],
               })
    obj = Mock(_rlc=rlc)
    rlc.attributes.return_value = ['id', 'attr1', 'somedate', 'somelist']
    rlc.relations.return_value = ['rel1', 'rel2']
    return obj


def test_it_can_serialize_an_object(s):
    obj = setup_obj()
    obj._dirty = set(['attr1', 'rel1'])
    result = s.serialize(obj)
    assert result == {
        'attr1': 'someattr',
        'somedate': '2018-01-01T00:00:00',
        'somelist': [1, 2, 3],
        'rel1': {
            'id': 2
        },
        'rel2': [{
            'id': 3
        }]
    }


def test_it_can_serialize_a_dirty_object(s):
    obj = setup_obj()
    obj._rlc.dirty = set(['attr1', 'rel1'])
    result = s.serialize_dirty(obj)
    assert result == {'attr1': 'someattr', 'rel1': {'id': 2}}



import os
from functools import partial

import requests


class AuthenticationError(Exception):
    pass


class BaseSession(requests.Session):
    def __init__(self,
                 url=None,
                 username=None,
                 password=None,
                 token=None,
                 verify=True,
                 trust_env=True,
                 **kwargs):
        super().__init__()
        self.verify = verify
        self.trust_env = trust_env
        self.kwargs = kwargs

        if url and username:
            self.authenticate(url, username, password)
        elif url and token:
            self.authenticate(url, token)

    def authenticate(self, url, username, password=None):
        raise NotImplementedError()

    def request(self, *args, **kwargs):
        return self.validate_response(super().request(*args, **kwargs))

    def validate_response(self, res):
        # raise an exception if status is 400 or up
        try:
            json_data = res.json()
        except Exception:
            json_data = {
                'message': 'Unspecified error ({})'.format(res.content)
            }
        # prepare error message
        res.reason = "{} ({})".format(res.reason, json_data)
        # raise if needed
        res.raise_for_status()
        return res


class BasicAuthSession(BaseSession):
    def authenticate(self, url, username, password):
        self.request = partial(self.request, auth=(username, password))


class BearerSession(BaseSession):
    def __init__(self, *args, **kwargs):
        self.bearer_header = kwargs.pop('bearer_header', 'Authorization')
        self.bearer_prefix = kwargs.pop('bearer_prefix', 'Bearer')
        super().__init__(*args, **kwargs)

    def authenticate(self, url, username, password=None):
        payload = {
            self.kwargs.get('user_field', 'username'): username,
            self.kwargs.get('password_field', 'password'): password
        }
        r = self.post(url, data=payload)
        if not r.ok:
            raise AuthenticationError(r.content)
        try:
            token = r.json().get('access_token')
        except Exception as e:
            raise AuthenticationError(
                "An error occured when authenticating".format(e))

        if not token:
            msg = ('An error occurred when authenticating: %s' % r.json())
            AuthenticationError(msg)
        self.headers.update(
            {self.bearer_header: '{} {}'.format(self.bearer_prefix, token)})


def Session(*args, **kwargs):
    auth_types = {
        'basic': BasicAuthSession,
        'bearer': BearerSession,
    }
    config = {}
    for key, value in os.environ.items():
        if key.startswith('RESTLESS_CLIENT_'):
            config_key = key.replace('RESTLESS_CLIENT_', '').lower()
            config[config_key] = value

    SessionClass = auth_types[config.pop('auth_type', 'bearer')]
    kwargs.update(config)
    return SessionClass(*args, **kwargs)



import json
import logging

from .inspect import inspect

logger = logging.getLogger('restless-client')

INVERT_OPERATORS = [('==', '!='), ('>', '<='), ('<', '>='), ('>=', '<'),
                    ('<=', '>'), ('in', 'not_in'), ('is_null', 'is_not_null')]
INVERT_MAPPING = {}
for a, b in INVERT_OPERATORS:
    INVERT_MAPPING[a] = b
    INVERT_MAPPING[b] = a


def is_filter_result(f, operator):
    assert isinstance(
        f,
        (BooleanResult, ComparisonResult)), 'Wrong use of {}'.format(operator)


class FilterCollection(list):
    def to_raw_filter(self):
        return [i.to_raw_filter() for i in self]


class BooleanResult:
    def __init__(self, operator, collection):
        self.operator = operator
        assert isinstance(collection, FilterCollection)
        self.collection = collection

    def __and__(self, other):
        return self.__boolean_operator('and', other)

    def __or__(self, other):
        return self.__boolean_operator('or', other)

    def __boolean_operator(self, operator, other):
        is_filter_result(self, operator)
        is_filter_result(other, operator)
        if self.operator == operator:
            self.collection.append(other)
            return self
        return BooleanResult(operator, FilterCollection([self, other]))

    def to_raw_filter(self):
        return {self.operator: [c.to_raw_filter() for c in self.collection]}

    def __str__(self):
        return str(self.to_raw_filter())


class ComparisonResult:
    def __init__(self, name, op, val):
        self.name = name
        self.op = op
        self.val = val

    def __invert__(self):
        if self.op not in INVERT_MAPPING:
            raise Exception('{} is not a valid invert candidate'.format(
                self.op))
        return ComparisonResult(self.name, INVERT_MAPPING[self.op], self.val)

    def __and__(self, other):
        return self.__boolean_operator('and', other)

    def __or__(self, other):
        return self.__boolean_operator('or', other)

    def __boolean_operator(self, operator, other):
        is_filter_result(self, operator)
        is_filter_result(other, operator)
        if isinstance(other, BooleanResult) and other.operator == operator:
            other.collection.append(self)
            return other
        return BooleanResult(operator, FilterCollection([self, other]))

    def to_raw_filter(self):
        return {"name": self.name, "op": self.op, "val": self.val}

    def __str__(self):
        return str(self.to_raw_filter())


class FilterMixIn:
    def __eq__(self, other):
        return self._assemble_filter('==', other)

    def __ne__(self, other):
        return self._assemble_filter('!=', other)

    def __gt__(self, other):
        return self._assemble_filter('>', other)

    def __ge__(self, other):
        return self._assemble_filter('>=', other)

    def __lt__(self, other):
        return self._assemble_filter('<', other)

    def __le__(self, other):
        return self._assemble_filter('<=', other)

    def in_(self, items):
        return self._assemble_filter('in', items)

    def has_(self, comparison_result):
        assert isinstance(comparison_result, ComparisonResult)
        return self._assemble_filter('has', comparison_result.to_raw_filter())

    def any_(self, comparison_result):
        assert isinstance(comparison_result, ComparisonResult)
        return self._assemble_filter('any', comparison_result.to_raw_filter())

    def like_(self, other):
        return self._assemble_filter('like', other)

    def _assemble_filter(self, op, val):
        return ComparisonResult(self.attribute, op, self._clean(val))

    def _clean(self, val):
        return val


class QueryFactory:
    def __init__(self, connection, cls):
        self.connection = connection
        self.cls = cls

    def __get__(self, obj, objtype=None):
        if obj:
            raise AttributeError('Cannot call query on from an instance')
        return Query(self.connection, self.cls)

    def __set__(self, obj, value):
        raise ValueError('Cannot set query')


class Query:
    def __init__(self, connection, cls):
        self.connection = connection
        self.cls = cls
        self._query = {}

    def filter(self, *queries):  # noqa A003
        q = []
        for query in queries:
            is_filter_result(query, 'filter')
            q.append(query.to_raw_filter())
        if not self._query.get('filters'):
            self._query['filters'] = []
        self._query['filters'].extend(q)
        return self

    def filter_by(self, **kwargs):
        filters = []  # noqa F841
        for attr, value in kwargs.items():
            self.filter(getattr(self.cls, attr) == value)
        return self

    def limit(self, limit):
        assert int(limit)
        self._query['limit'] = limit
        return self

    def offset(self, offset):
        assert int(offset)
        self._query['offset'] = offset
        return self

    def order_by(self, **kwargs):
        order_by = []
        for attr, direction in kwargs.items():
            order_by.append({"field": attr, "direction": direction})
        if order_by:
            self._query['order_by'] = order_by
        return self

    def group_by(self, *group_by):
        if group_by:
            self._query['group_by'] = group_by
        return self

    def first(self):
        self.limit(1)
        self.order_by(**{self.cls._rlc.pk_name: 'asc'})
        return self.one()

    def last(self):
        self.limit(1)
        self.order_by(**{self.cls._rlc.pk_name: 'desc'})
        return self.one()

    def one(self):
        self._query['single'] = True
        kwargs = {'single': True}
        if self._query:
            kwargs['q'] = self._get_query()
        return self.connection.load_query(self.cls, **kwargs)

    def one_or_none(self):
        try:
            return self.one()
        except Exception as e:
            if 'Multiple results found' in str(e):
                raise e

    def all(self):  # noqa A003
        kwargs = {}
        if self._query:
            kwargs['q'] = self._get_query()
        return self.connection.load_query(self.cls, **kwargs)

    def get(self, oid):
        registry_id = '{}{}'.format(self.cls.__name__, oid)
        meta = inspect(self.cls)
        if registry_id in meta.client.registry:
            return meta.client.registry[registry_id]
        return self.connection.load(self.cls, oid)

    def _get_query(self):
        return json.dumps(self._query)



from collections import namedtuple

import pytest

from restless_client.connection import Connection
from restless_client.filter import Query
from restless_client.inspect import inspect


class Meta:
    def __init__(self, base_url, pk_name, client):
        self.base_url = base_url
        self.pk_name = pk_name
        self.client = client


class BaseObject:
    _rlc = Meta(
        base_url="http://app/api/formicarium",
        pk_name="id",
        client=None,
    )

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)


class CollectionClass(list):
    def __init__(self, collection_class, objs):
        super().__init__(objs)


@pytest.fixture
def query(cl):
    cl.opts.CollectionClass = CollectionClass
    inspect(BaseObject).client = cl
    return Query(Connection(cl.opts.session, cl.opts), BaseObject)


def test_it_can_perform_an_all(query):
    result = query.all()
    assert len(result) == 5


def test_it_can_perform_an_all_on_several_pages(session, instances, query):
    for i in range(100):
        session.add(instances.Formicarium(name=f"MrNobody{i}"))
    session.commit()
    result = query.all()
    assert len(result) == 105


def test_it_can_perform_a_filter(query, cl):
    expected = 'Specimen-1'
    result = query.filter(cl.Formicarium.name == expected).one()
    assert result.name == expected


def test_it_can_perform_a_filter_by(query, cl):
    expected = 'Specimen-1'
    query.cls.name = cl.Formicarium.name
    result = query.filter_by(name=expected).one()
    assert result.name == expected


def test_it_can_perform_a_limit(query):
    result = query.limit(3).all()
    assert len(result) == 3


def test_it_can_perform_an_offset(query):
    result = query.offset(2).all()
    assert len(result) == 3


def test_it_can_perform_an_order_by_asc(query):
    result = query.order_by(name='asc').all()
    assert result[0].name == 'PAnts'


def test_it_can_perform_an_order_by_desc(query):
    result = query.order_by(name='desc').all()
    assert result[0].name == 'The yard yokels'


def test_it_can_perform_a_first(query):
    result = query.first()
    assert result.name == 'Specimen-1'


def test_it_can_perform_a_last(query):
    result = query.last()
    assert result.name == 'The Free SociAnty'


def test_it_can_perform_a_one_or_none_and_returns_none(query, cl):
    result = query.filter(
        cl.Formicarium.name == 'does not exist').one_or_none()
    assert result is None


def test_it_throws_an_error_when_one_or_none_returns_multiple_instances(query):
    with pytest.raises(Exception):
        query.one_or_none()


def test_it_can_perform_a_get(query):
    result = query.get(1)
    assert result.id == 1


def test_query_does_not_set_attributs_as_dirty(fcl):
    o = fcl.Object3
    results = o.query.all()
    assert not any([inspect(res).dirty for res in results])


def test_chaining_query_with_filter_does_not_have_side_effects(fcl):
    o = fcl.Object1
    objects = o.query.all()
    assert len(objects) == 5
    assert len(o.query.filter(o.attribute1 == "o1a11").all()) == 1
    assert len(o.query.all()) == 5



import logging
import pprint
from functools import partial, wraps

import requests
from ordered_set import OrderedSet
from requests.packages.urllib3.exceptions import InsecureRequestWarning

from .utils import parse_custom_types, urljoin

requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
logger = logging.getLogger('restless-client')


def get_url(obj):
    return obj._rlc.base_url


def log(fn):
    @wraps(fn)
    def decorator(*args, **kwargs):
        logger.debug('kwargs: {}'.format(pprint.pformat(kwargs)))
        res = fn(*args, **kwargs)
        logger.debug('result: {}'.format(pprint.pformat(res)))
        return res

    return decorator


def lock_loading(fn):
    @wraps(fn)
    def decorator(self, obj, *args, **kwargs):
        with obj._rlc.client.loading:
            return fn(self, obj, *args, **kwargs)

    return decorator


def raise_on_locked(fn):
    def decorator(self, obj, *args, **kwargs):
        """
        mostly a utility and debugging measure. We'd like to prevent a load from
        being triggered when another load is in progress. Throwing a hard
        exception will give us a handle on the problem much faster.
        """
        client = obj._rlc.client
        if client.is_loading and client.opts.debug:
            raise Exception('Loading is locked')
        return fn(self, obj, *args, **kwargs)

    return decorator


class Connection:
    def __init__(self, client, opts):
        self.client = client
        self.session = opts.session
        self.opts = opts

    @raise_on_locked
    @lock_loading
    def load_query(self, obj_class, single=False, **kwargs):
        raw = self.request(obj_class._rlc.base_url, params=kwargs)

        if single:
            return obj_class(**raw)

        # iterate over pages
        objects = raw['objects']
        for page in range(2, raw['total_pages'] + 1):
            kwargs['page'] = page
            objects.extend(
                self.request(obj_class._rlc.base_url,
                             params=kwargs)['objects'])

        return self.opts.CollectionClass(
            obj_class, OrderedSet([obj_class(**obj) for obj in objects]))

    @raise_on_locked
    @lock_loading
    def load(self, obj_class, obj_id):
        raw = self.request(urljoin(obj_class._rlc.base_url, str(obj_id)))
        return obj_class(**raw)

    @lock_loading
    def create(self, obj, object_dict=None):
        object_dict = object_dict or self.client.serializer.serialize_dirty(
            obj)

        property_values = {}
        for property_name in obj._rlc.dirty_properties:
            property_values[property_name] = object_dict.pop(property_name)

        url = obj._rlc.base_url
        r = self.request(url, http_method='post', json=object_dict)
        setattr(obj, obj._rlc.pk_name, r[obj._rlc.pk_name])

    @lock_loading
    def update(self, obj, object_dict=None):
        object_dict = object_dict or self.client.serializer.serialize_dirty(
            obj)
        url = urljoin(obj._rlc.base_url, str(obj._rlc.pk_val))
        self.request(url, http_method='put', json=object_dict)

    def delete(self, obj):
        if not obj._rlc.is_new:
            url = urljoin(obj._rlc.base_url, str(obj._rlc.pk_val))
            self.request(url, http_method='delete')

    @log
    def request(self, url, **kwargs):
        method = kwargs.pop('http_method', 'get')
        fn = getattr(self.session, method)
        r = fn(url, **kwargs)
        if method == 'delete':
            return

        result = r.json(object_hook=partial(parse_custom_types, **kwargs), )
        return result



import pytest

from restless_client import inspect


def test_it_can_load_an_object(cl):
    colony = cl.AntColony.query.get(1)
    assert colony.name == 'Argentine Ant'


def test_it_can_load_a_simple_filter(cl):
    colony = cl.AntColony.query.filter_by(name='Argentine Ant').one()
    assert colony.name == 'Argentine Ant'


def test_it_cant_set_an_unexisting_attribute(cl):
    colony = cl.AntColony.query.get(1)
    with pytest.raises(AttributeError):
        colony.unknown = 'Argentine Ant'


def test_it_can_load_a_simple_relation_filter(cl):
    ofilter = cl.AntColony.formicarium.name == 'Specimen-1'
    colony = cl.AntColony.query.filter(ofilter).one()
    assert colony.name == 'Argentine Ant'


def test_it_can_load_based_on_an_object_filter(cl):
    formicarium = cl.Formicarium.query.filter_by(name='Specimen-1').one()
    colony = cl.AntColony.query.filter_by(formicarium=formicarium).one()
    assert colony.name == 'Argentine Ant'


def test_it_can_init_a_polymorphed_class_correcly(cl):
    formicarium = cl.Formicarium.query.filter_by(name='Specimen-1').one()
    assert isinstance(formicarium, cl.SandwichFormicarium)
    expected = [
        'id', 'name', 'formicarium_type', 'width', 'collection_id', 'height'
    ]
    assert inspect(formicarium).attributes() == expected
    expected = ['collection', 'colonies']
    assert inspect(formicarium).relations() == expected
    assert formicarium.height == 10


def test_it_can_load_an_inherited_object(cl):
    formicarium = cl.SandwichFormicarium.query.filter_by(
        name='Specimen-1').one()
    assert formicarium.height == 10


def test_it_correctly_loads_inherited_objects(cl):
    formicaria = cl.Formicarium.query.all()
    classes = set([inspect(f).class_name for f in formicaria])
    assert set(['SandwichFormicarium', 'FreeStandingFormicarium']) == classes


def test_it_can_save_an_object(cl):
    collection = cl.AntCollection(name="Antiquities", location="The Past")
    cl.save(collection)
    assert collection.id == 4


def test_it_can_save_an_inherited_object(cl):
    formicarium = cl.SandwichFormicarium(
        name="PlAnts",
        height=10,
        width=10,
        collection=cl.AntCollection.query.get(1),
    )
    cl.save(formicarium)
    assert formicarium.id == 6


def test_it_can_save_an_object_with_an_unsaved_object_as_relation(cl):
    collection = cl.AntCollection(name="Antiquities", location="The Past")
    formicarium = cl.SandwichFormicarium(
        name="PlAnts",
        height=10,
        width=10,
        collection=collection,
    )
    cl.save(formicarium)
    assert collection.id == 4
    assert formicarium.id == 6


def test_it_can_save_an_object_with_unsaved_objects_as_relation(cl):
    formicarium = cl.SandwichFormicarium(
        name="PlAnts",
        height=10,
        width=10,
    )
    collection = cl.AntCollection(name="Antiquities",
                                  location="The Past",
                                  formicaria=[formicarium])
    cl.save(collection)
    assert collection.id == 4
    assert formicarium.id == 6


def test_it_can_save_an_object_when_appending_unsaved_objects_as_relation(cl):
    formicarium = cl.SandwichFormicarium(
        name="PlAnts",
        height=10,
        width=10,
    )
    collection = cl.AntCollection(
        name="Antiquities",
        location="The Past",
    )
    collection.formicaria.append(formicarium)
    cl.save(collection)
    assert collection.id == 4
    assert formicarium.id == 6


def test_it_can_update_an_object(cl, app):
    new_name = "ElephAnt"
    collection = cl.AntCollection.query.get(1)
    collection.name = new_name
    cl.save(collection)
    assert app.AntCollection.query.get(1).name == new_name


def test_it_can_update_an_object_when_removing_and_object_from_relations(
        cl, app):
    formicarium = cl.Formicarium.query.get(1)
    formicarium.colonies.remove(formicarium.colonies[0])
    cl.save(formicarium)
    assert app.Formicarium.query.get(1).colonies == []


def test_it_can_delete_an_object(cl, app):
    cl.delete(cl.AntColony.query.get(1))
    assert app.AntColony.query.get(1) is None


def test_it_can_access_a_level_two_relation(cl):
    colony = cl.AntColony.query.get(1)
    collection = colony.formicarium.collection
    assert collection.name == 'Antopia'


def test_it_can_filter_a_level_two_relation_up(cl):
    result = cl.AntColony.query.filter(
        cl.AntColony.formicarium.collection.name == 'Antics').all()
    expected = ['Fire Ant', 'Garden Ant', 'Bulldog Ant']
    assert sorted([r.name for r in result]) == sorted(expected)


def test_it_can_filter_a_level_two_relation_down(cl):
    result = cl.AntCollection.query.filter(
        cl.AntCollection.formicaria.colonies.name == 'Argentine Ant').all()
    assert result.one().name == 'Antopia'


def test_it_can_filter_boolean_expressions(cl):
    result = cl.AntColony.query.filter(
        ((cl.AntColony.name == 'Argentine Ant')
         | (cl.AntColony.queen_size > 15))
        & cl.AntColony.color.in_(['red', 'brown'])).all()
    expected = ['Fire Ant', 'Argentine Ant', 'Bulldog Ant']
    assert sorted([r.name for r in result]) == sorted(expected)


# tests not suited for this module, need to be moved



import logging

from prettytable import PrettyTable

logger = logging.getLogger('restless-client')


def update_backref(remove):
    def outer_decorator(fn):
        def decorator(self, *args, **kwargs):
            res = fn(self, *args, **kwargs)
            item = res or args[0]
            if not self.parent._rlc.client.is_loading:
                self.parent._rlc.dirty.add(self.for_attr)
            if self.for_attr:
                self._update_backref(item, self.for_attr, remove=remove)
            return res

        return decorator

    return outer_decorator


class TypedList(list):
    def __init__(self, otype, parent, for_attr=None):
        self.type = otype
        self.parent = parent
        self.for_attr = for_attr

    @update_backref(remove=False)
    def append(self, item):
        if not isinstance(item, self.type):
            cls_name = item.__class__.__name__
            msg = 'Only {} can be added, {} provided'
            raise TypeError(msg.format(self.type.__name__, cls_name))
        list.append(self, item)

    def extend(self, lst):
        for i in lst:
            self.append(i)

    @update_backref(remove=True)
    def remove(self, item):
        list.remove(self, item)

    @update_backref(remove=True)
    def pop(self):
        return list.pop(self)

    def _update_backref(self, item, attr, remove=False):
        relhelper = self.parent._rlc.relhelper
        backref = relhelper.backref(attr)
        if backref:
            msg = 'Updating {} backref {}.{}'.format(
                item._rlc.relhelper.type(backref), item, backref)
            if item._rlc.relhelper.is_scalar(backref):
                value = None if remove else self.parent
                logger.debug(msg)
                with self.parent._rlc.client.loading:
                    setattr(item, backref, value)
            else:
                if backref in item._rlc.values:
                    fn = list.remove if remove else list.append
                    lst = getattr(item, backref)
                    if (self.parent not in lst) ^ remove:
                        logger.debug(msg)
                        fn(lst, self.parent)


# only used for printing puroposes, has no functional benefit
class ObjectCollection(list):
    def __init__(self, object_class, lst=None, attrs=None):
        self.object_class = object_class
        self.attrs = attrs or object_class._rlc.attributes()
        if lst:
            self.extend(lst)

    def first(self):
        return self[0]

    def one(self):
        if len(self) > 1:
            raise ValueError('more than one result')
        return self.first()

    def __getitem__(self, key):
        if isinstance(key, int):
            return list.__getitem__(self, key)
        if isinstance(key, str):
            key = [key]
        return ObjectCollection(self.object_class, self, attrs=key)

    def pprint(self):
        attrs = self.attrs
        # make sure id and name are at the front of the table
        headers = ['id']
        if 'name' in attrs:
            headers.append('name')
        columns = set(attrs).difference(set(['id', 'name']))
        relation_columns = [c for c in columns if str(c).endswith('_id')]
        data_columns = columns.difference(set(relation_columns))
        # make sure to show the data columns next
        headers.extend(sorted(data_columns))
        # then show relation columns
        headers.extend(sorted(relation_columns))
        pt = PrettyTable(headers)
        pt.align = 'l'

        def truncate(val):
            val = val.replace('\n', '')
            if len(val) > 25:
                val = '{}...'.format(val[:22])
            return val

        for obj in self:
            values = obj._values
            pt.add_row(
                [truncate(str(values.get(header))) for header in headers])
        res = pt.get_string(border=False, sortby="id")
        if not res:
            pt.add_row(['' for header in headers])
            res = pt.get_string(border=False, sortby="id")
        return res

    def __repr__(self):
        return self.pprint()



import inspect as p_inspect
import logging
from itertools import chain

logger = logging.getLogger('restless-client')


class ModelMeta:
    def __init__(self, client, class_name, pk_name, attributes, properties,
                 relations, methods, base_url, method_url, property_url,
                 polymorphic, relhelper):
        self.client = client
        self.connection = client.connection

        self.base_url = base_url
        self.method_url = method_url
        self.property_url = property_url

        self.class_name = class_name
        self.pk_name = pk_name

        self.properties = properties
        self._attributes = attributes
        self._relations = relations
        self._methods = methods

        self.deserializer = client.deserializer
        self.serializer = client.serializer
        self.polymorphic = polymorphic
        self.relhelper = relhelper

    def attributes(self):
        return list(self._attributes.keys())

    def relations(self):
        return list(self._relations.keys())

    def methods(self):
        return list(self._methods)


class InstanceState:
    def __init__(self, instance):
        self.instance = instance
        self.meta = instance._rlc
        self.dirty = set()
        self.values = {}

    def __getattribute__(self, attrib):
        meta = super().__getattribute__('meta')
        if hasattr(meta, attrib):
            return getattr(meta, attrib)
        return super().__getattribute__(attrib)

    @property
    def pk_val(self):
        return getattr(self.instance, self.meta.pk_name)

    @property
    def is_new(self):
        return str(self.pk_val).startswith('C')

    @property
    def settable_attributes(self):
        settable_props = [d for (d, s) in self.properties.items() if s]
        return chain(self._relations, self._attributes, settable_props)

    @property
    def dirty_properties(self):
        props = set()
        for property_name in self.meta.properties:
            if property_name in self.dirty:
                props.add(property_name)
        return props

    def delete(self):
        self.connection.delete(self.instance)

    def save(self):
        if self.is_new:
            self.connection.create(self.instance)
        elif self.dirty:
            self.connection.update(self.instance)
        else:
            logger.debug("No action needed")
        self.dirty = set()


def inspect(obj):
    meta = obj._rlc
    if isinstance(meta, InstanceState):
        return meta
    if p_inspect.isclass(obj):
        return meta
    return InstanceState(obj)



from unittest.mock import patch

import pytest
import requests

from restless_client.utils import State


@patch("restless_client.connection.Connection.request")
def test_it_set_status_back_to_loadable_if_httperror(request, mcl):
    request.side_effect = requests.HTTPError("Some issue")
    with pytest.raises(requests.HTTPError):
        apt = mcl.Apartment.query.get(1)
    assert mcl.state == State.LOADABLE



from collections import defaultdict
from datetime import date, datetime, timedelta

from restless_client import types
from restless_client.inspect import inspect


def test_it_can_run_a_remote_method_without_params(mcl):
    res = mcl.Apartment.query.one().function_without_params()
    assert res == 5


def test_it_can_run_a_remote_method_with_builtin_type_params(mcl):
    res = mcl.Apartment.query.one().function_with_params(5, 'test')
    assert res == '5: test'


def test_it_can_run_a_remote_method_with_kwargs(mcl):
    res = mcl.Apartment.query.one().function_with_kwargs(kwarg2=5,
                                                         kwarg1='test')
    assert res == 'test: 5'


def test_it_can_run_a_remote_method_with_args_kwargs(mcl):
    res = mcl.Apartment.query.one().funtion_with_args_kwargs(arg1=5,
                                                             kwarg1='test')
    assert res == '5: test'


def test_it_can_run_a_remote_method_with_default_params(mcl):
    res = mcl.Apartment.query.one().function_with_default_params(param2='test')
    assert res == '5: test'


def test_it_can_run_a_remote_method_with_an_object_as_param(mcl):
    colony = mcl.AntColony.query.get(1)
    res = mcl.Apartment.query.one().function_with_an_object(colony)
    assert isinstance(res, mcl.AntColony)
    assert res.name == colony.name


def test_it_can_run_a_remote_method_returning_uncommitted_obj(mcl):
    res = mcl.Apartment.query.one().function_with_uncommitted_object()
    assert isinstance(res, mcl.Apartment)
    assert res.name == 'Oof-Owie'
    assert not inspect(res).pk_val


def test_it_can_get_a_remote_property(mcl):
    assert mcl.Apartment.query.one().some_property == 'a_property_value'


def test_it_can_set_a_remote_property(mcl):
    apt = mcl.Apartment.query.one()
    assert apt.settable_property == 'ApAntMent'
    apt.settable_property = 'ApSetMent'
    mcl.save(apt)
    assert mcl.Apartment.query.one().settable_property == 'ApSetMent'


def test_it_casts_attribute_types_correctly_when_not_marshalling_from_json(
        mcl):
    apt = mcl.Apartment.query.one()
    mt = apt.function_with_new_obj()

    expected_types = {
        "date": date,
        "dt": datetime,
        "json": dict,
    }
    for key, type_ in expected_types.items():
        assert type(getattr(mt, key)) == type_


def test_it_can_execute_an_object_hook(mcl):
    types.OBJECT_HOOKS = defaultdict(dict)

    @types.object_hook(mcl.Apartment.some_hybrid)
    def hybrid_parser(value):
        return mcl.AntCollection(**value)

    apt = mcl.Apartment.query.one()
    assert apt.some_hybrid.__class__ == mcl.AntCollection


def test_it_can_execute_an_object_hook_defined_as_strings(mcl):
    types.OBJECT_HOOKS = defaultdict(dict)

    @types.object_hook('Apartment', 'some_hybrid')
    def hybrid_parser(value):
        return mcl.AntCollection(**value)

    apt = mcl.Apartment.query.one()
    assert apt.some_hybrid.__class__ == mcl.AntCollection



import json
from collections import defaultdict

from dateutil import parser

CASTERS = {}
OBJECT_HOOKS = defaultdict(dict)


def register_type(type_name, caster):
    CASTERS[type_name] = caster


def cast_type(value, desired_type, raise_on_error=False):
    if desired_type in CASTERS:
        try:
            return CASTERS[desired_type](value)
        except Exception as e:
            if raise_on_error:
                raise e
    return value


def date_caster(value):
    return parser.parse(value).date()


def datetime_caster(value):
    return parser.parse(value)


def boolean_caster(value):
    if value in ('False', 'false', '0', 'no'):
        return False
    bool(value)


def json_caster(value):
    return json.loads(value)


register_type('date', date_caster)
register_type('datetime', datetime_caster)
register_type('utcdatetime', datetime_caster)
register_type('boolean', boolean_caster)
register_type('json', json_caster)
register_type('jsontype', json_caster)


def object_hook(model, attribute=None):
    if not isinstance(model, str):
        attribute = model.parent_attribute
        model = model.parent_klass.__name__

    def outer_decorator(fn):
        OBJECT_HOOKS[model][attribute] = fn
        return fn

    return outer_decorator


def object_hook_emit(model, attribute, value):
    if model.__name__ in OBJECT_HOOKS:
        fn = OBJECT_HOOKS[model.__name__].get(attribute)
        if fn:
            return fn(value)
    return value



import inspect

from .utils import State


class UncallableMethod(Exception):
    pass


def construct_method(opts, client, method, method_details):
    method = opts.Method(method, method_details, client.connection)

    def func(self, *args, **kwargs):
        return method(self, *args, **kwargs)

    # code = func.__code__
    # func = types.FunctionType(code, func.__globals__, method.name)
    params = []
    for param in method_details['args']:
        params.append(
            inspect.Parameter(param, inspect.Parameter.POSITIONAL_OR_KEYWORD))

    for param in method_details['kwargs']:
        params.append(
            inspect.Parameter(param,
                              inspect.Parameter.POSITIONAL_OR_KEYWORD,
                              default=State.VOID))

    func.__signature__ = inspect.Signature(params)
    func.__name__ = method.name
    return func


class Method:
    def __init__(self, name, details, connection):
        self.name = name
        self.connection = connection
        self.args = details['args']
        self.kwargs = details['kwargs']
        self.argsvar = details['argsvar']
        self.kwargsvar = details['kwargsvar']

    def __call__(self, obj, *args, **kwargs):
        if obj._rlc.is_new:
            raise UncallableMethod("Cannot call methods on new objects.")
        self.validate_params(args, kwargs)
        rlc = obj._rlc
        url = '{}/{}/{}'.format(rlc.method_url, rlc.pk_val, self.name)
        payload = {'payload': self.serialize_params(args, kwargs)}
        result = self.connection.request(url, http_method='post', json=payload)
        result = self.cereal.loads(result['payload'])
        return result

    @property
    def cereal(self):
        return self.connection.client.cereal

    def serialize_params(self, args, kwargs):
        return self.cereal.dumps({'args': args, 'kwargs': kwargs})

    def validate_params(self, args, kwargs):
        if not self.argsvar and len(args) < len(self.args):
            msg = '{}() missing {} required positional argument: {}'
            diff = self.args[len(args):]
            TypeError(msg.format(self.name, len(diff), ', '.join(diff)))
        kwdiff = set(self.kwargs).difference(kwargs.keys())
        if not self.kwargsvar and kwdiff:
            msg = "{}() got an unexpected keyword argument '{}'"
            TypeError(msg.format(self.name, kwdiff.pop()))

        for key, val in list(kwargs.items()):
            if val == State.VOID:
                kwargs.pop(key)
